/**
This code is part of the Arc-flow 0-1CSP Solver.

Copyright (C) 2013, Filipe Brandao
Faculdade de Ciencias, Universidade do Porto
Porto, Portugal. All rights reserved. E-mail: <fdabrandao@dcc.fc.up.pt>.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#ifndef _CSP01FLOW_HPP_
#define _CSP01FLOW_HPP_

#include <map>
#include <set>
#include <vector>
#include <cmath>
#include <ctime>
#include <cassert>
#include <climits>
#include <algorithm>
#include "gurobi_c.h"
#include "gurobi_c++.h"
using namespace std;

typedef pair<int,int> int_pair;

#define EPS 1e-5
#define MAX_LEN 256
#define TRANSPOSE (1)
#define MP(x,y) make_pair(x,y)
#define All(x) (x).begin(),(x).end()
#define ForEach(i,c) for(__typeof((c).begin()) i = (c).begin(); i != (c).end(); i++)
#define CURTIME clock()
#define TIMEDIF(t0) (double(clock()-t0)/CLOCKS_PER_SEC)


class Item{
public:
    int w;
    int b;
    int id;
    bool operator<(const Item &o) const{
        return w < o.w || (w == o.w && b == o.b);
    }
};


class Instance{
public:
    int W, m;
    vector<Item> items;
};


class NodeSet{
private:
    map<int_pair, int> index;
    vector<int_pair> labels;
public:
    int get_index(const int_pair&lbl){    
        map<int_pair, int>::iterator itr = index.find(lbl);
        if(itr != index.end()){
            return itr->second;
        }else{
            int ind = labels.size();
            labels.push_back(lbl);
            index[lbl] = ind;
            return ind;   
        }
    }
    
    int_pair get_label(int ind) const {
        assert(ind < (int)labels.size());
        return labels[ind];
    }
    
    int size() const {
        return labels.size();
    }   
    
    void clear(){
        index.clear();
        labels.clear();
    }       
    
    void sort(){
        index.clear();
        ::sort(All(labels));
        int pos = 0;        
        ForEach(itr, labels)
            index[*itr] = pos++;
    }   
    
    vector<int> topological_order() const {
        vector<pair<int_pair, int> > tmp(All(index));
        vector<int> ord(tmp.size());
        int pos = 0;
        ForEach(itr, tmp)
            ord[itr->second] = pos++;
        return ord;        
    }
};


class Arc{
public:
    int u;
    int v;
    int label;
    Arc(int _u = -1, int _v = -1, int _label = -1): 
        u(_u), v(_v), label(_label){}
    bool operator<(const Arc &o) const{
        return (u < o.u) || 
               (u == o.u && v < o.v) || 
               (u == o.u && v == o.v && label < o.label);
    }
    bool operator==(const Arc &o) const{
        return u == o.u && v == o.v && label == o.label;
    }  
};

vector<vector<int_pair> > get_adj(int nv, const vector<Arc> &arcs, 
        bool transpose = false){
    vector<vector<int_pair> > adj(nv);    
    ForEach(itr, arcs){
        const int u = itr->u;
        const int v = itr->v;  
        assert(u < nv && v < nv);
        if(!transpose)
            adj[u].push_back(make_pair(v, itr->label));
        else
            adj[v].push_back(make_pair(u, itr->label));
    }    
    return adj;
}

class AFGraph{
public:
    int nv;
    vector<Arc> arcs;    
    AFGraph(int _nv, const vector<Arc> &_arcs): nv(_nv), arcs(_arcs){}
    vector<vector<int_pair> > get_adj(bool transpose = false){
        return ::get_adj(nv, arcs, transpose);
    }
};

class ArcflowSol{
private:
    map<Arc, int> flow;     
    int S, T;
    
    vector<pair<int, vector<int_pair> > > remove_excess(
        const vector<pair<int, vector<int> > > &sol, vector<int> dem) const{    
        vector<pair<int, vector<int_pair> > > tmp;
        ForEach(pat, sol){
            map<int, int> count;
            ForEach(it, pat->second) count[*it] += 1;
            vector<int> rm;
            int rep = pat->first;
            while(rep > 0){
                rm.clear();
                ForEach(itr, count){           
                    itr->second = min(itr->second, dem[itr->first]);
                    if(itr->second == 0) rm.push_back(itr->first);
                }
                ForEach(ind, rm) count.erase(*ind);
                            
                int f = rep;
                ForEach(itr, count)
                    f = min(f, dem[itr->first]/itr->second);
                rep -= f;
                
                tmp.push_back(make_pair(f, vector<int_pair>(All(count))));            
                ForEach(itr, count)           
                    dem[itr->first] -= f * itr->second;
            }
        }

        map<vector<int_pair>, int> m;
        ForEach(itr, tmp){
            sort(All(itr->second));
            m[itr->second] += itr->first;
        }
        
        vector<pair<int, vector<int_pair> > > final;
        ForEach(itr, m)
            final.push_back(make_pair(itr->second, itr->first));
        return final;
    }
        
    bool is_valid(vector<pair<int, vector<int_pair> > > sol,
        const Instance &inst) const{
        vector<int> dem(inst.m);
        for(int i = 0; i < inst.m; i++)
            dem[i] = inst.items[i].b;
        ForEach(pat, sol){
            int w = 0;
            int prev = -1;
            ForEach(itr, pat->second){
                if(itr->first == prev) return false;            
                w += inst.items[itr->first].w;
                dem[itr->first] -= pat->first * itr->second;            
                prev = itr->first;                
            }
            if(w > inst.W) return false;
        }
        for(int i = 0; i < inst.m; i++){
            if(dem[i] != 0) return false;
        }
        return true;
    }      

public:
    ArcflowSol(const map<Arc, int> &_flow, int _S, int _T):
        flow(_flow), S(_S), T(_T){}

    vector<pair<int, vector<int_pair> > > extract_solution(
        const vector<int> &dem){
        set<int> nodes;
        map<int, vector<Arc> > adj;
        ForEach(a, flow){
            int u = a->first.u;
            int v = a->first.v;
            nodes.insert(u);
            nodes.insert(v); 
            adj[v].push_back(a->first);        
        }
            
        vector<int> lst(All(nodes));
        
        vector<pair<int, vector<int> > > sol;
        while(true){
            map<int, Arc> pred;
            map<int, int> dp;
            dp[S] = INT_MAX;    
            ForEach(v, lst){
                int &val = dp[*v];
                Arc &p = pred[*v];
                ForEach(a, adj[*v]){   
                    assert(dp.count(a->u) != 0);
                    int m = min(dp[a->u], flow[*a]);
                    if(m > val){
                        p = *a;
                        val = m;
                    }
                }
            }
            int f = dp[T];
            if(f == 0) break;                
            int v = T;
            sol.push_back(pair<int, vector<int> >());
            pair<int, vector<int> > &pat = sol.back();
            pat.first = f;
            while(v != S){
                Arc a = pred[v];
                int u = a.u;
                int lbl = a.label;
                if(lbl < (int)dem.size()) // != LOSS
                    pat.second.push_back(lbl);
                flow[a] -= f;
                v = u;
            }        
        }       
        int fs = 0;
        ForEach(a, flow) 
            fs += a->second;
        assert(fs == 0);         
        return remove_excess(sol, dem);
    }

        
    void print_solution(const Instance &inst, 
            bool print_inst = true, bool validate = true){    
        vector<int> dem(inst.m), id(inst.m), rid(inst.m);   
        for(int i = 0; i < inst.m; i++){
            dem[i] = inst.items[i].b;
            int j = inst.items[i].id;
            id[i] = j;
            rid[j] = i;
        }
        vector<pair<int, vector<int_pair> > > sol = extract_solution(dem);
        if(validate) 
            assert(is_valid(sol, inst));

        int obj = 0;
        ForEach(pat, sol)
            obj += pat->first;

        printf("Objective: %d\n", obj);

        printf("Solution:\n");
                   
        ForEach(pat, sol){       
            vector<int> tmp;
            ForEach(itr, pat->second){
                int t = id[itr->first]+1;   
                for(int i = 0; i < itr->second; i++)
                    tmp.push_back(t);
            }        
            sort(All(tmp));
            
            printf("%d x [", pat->first);
            ForEach(i, tmp){                        
                if(i != tmp.begin()) printf(", ");            
                printf("%d", *i);
            }         
            printf("]\n");
        }
        
        if(print_inst){
            printf("\nInstance:\n");
            for(int i = 0; i < inst.m; i++){
                int j = rid[i];
                printf("w_%d: %d, b_%d: %d;\n", i+1, inst.items[j].w, 
                                                i+1, inst.items[j].b);
            }                                           
        }
    }   
};

class Arcflow{
private:
    clock_t tstart;
    Instance inst;
    int W, m;
    vector<int> w;
    NodeSet NS;
    set<Arc> AS;
    vector<Arc> A;
    map<int_pair, int_pair> dp;

    void relabel_graph(const vector<int> &label){    
        set<Arc> arcs;
        ForEach(itr, A){
            int u = label[itr->u];
            int v = label[itr->v];
            if(u != v)
                arcs.insert(Arc(u, v, itr->label));
        }
        A.assign(All(arcs));
    }   
    
    int_pair go(const int_pair &u){
        map<int_pair, int_pair>::iterator itr = dp.find(u);        
        if(itr != dp.end())
            return itr->second;
        
        int it = u.second;
        int_pair mu(W, m);
        
        if(it == m)
            return mu;
                    
        int up = -1;
        if(it+1 < m){
            mu = go(MP(u.first, it+1));
            up = NS.get_index(mu);
        }
        
        if(u.first+w[it] <= W){
            int_pair v = go(MP(u.first+w[it], it+1));
            mu = MP(min(mu.first, v.first-w[it]), it);
            int iu = NS.get_index(mu);
            int iv = NS.get_index(v);            
            AS.insert(Arc(iu, iv, it));
            if(up != -1 && iu != up)
                AS.insert(Arc(iu, up, m));
        }

        dp[u] = mu;
        return mu;
    }
    
    void build(){
        dp.clear();
        AS.clear();        
        NS.clear();
        go(MP(0, 0));
        A.assign(All(AS));   
        printf("#dp: %d\n", (int)dp.size());
        dp.clear();
        AS.clear();         
    
        vector<int> order = NS.topological_order();
        relabel_graph(order);
        NS.sort();
    }
    
    void compress(){
        int nv = NS.size();
        vector<int> label(nv);
        vector<vector<int_pair> > adj = get_adj(nv, A, TRANSPOSE);
        
        NodeSet NStmp;    
        for(int u = 0; u < NS.size(); u++){
            int_pair lbl;
            ForEach(itr, adj[u]){            
                int v = label[itr->first];
                int it = itr->second;          
                int_pair lv = NStmp.get_label(v);
                lbl.first = max(lbl.first, lv.first+w[it]);
                lbl.second = max(lbl.second, lv.second);
                if(it != m)
                    lbl.second = max(lbl.second, it);
            }
            label[u] = NStmp.get_index(lbl);
        }

        NS = NStmp;    
        vector<int> order = NS.topological_order();
        ForEach(itr, label)
            *itr = order[*itr];
        relabel_graph(label);
        NS.sort();                
    }   
    
public:        
    Arcflow(const Instance &_inst, bool comp = true): inst(_inst){
        tstart = CURTIME;
        W = inst.W;
        m = inst.m;
        for(int i = 0; i < m; i++) 
            inst.items[i].id = i;        
        stable_sort(All(inst.items));
        reverse(All(inst.items));            
        
        w.resize(m);
        for(int i = 0; i < m; i++)
            w[i] = inst.items[i].w;
        w.push_back(0);
        
        build(); // build step-1 graph
        int nv1 = NS.size()+1;
        int na1 = A.size()+NS.size()-1;
        printf("Step-1 Graph: %d vertices and %d arcs (%.3fs)\n", 
            nv1, na1, TIMEDIF(tstart));
        
        if(comp){
            compress(); //build step-1 graph    
            int nv2 = NS.size()+1;
            int na2 = A.size()+NS.size()-1;
            printf("Step-2 Graph: %d vertices and %d arcs (%.3fs)\n", 
                nv2, na2, TIMEDIF(tstart));            
        }        
    }     
    
    void finalize(){
        int t = NS.size();
        for(int i = 1; i < NS.size(); i++)
            A.push_back(Arc(i, t, m));
    }         
    
    AFGraph get_graph() const {
        vector<Arc> At(A);
        ForEach(itr, At)
            if(itr->label != m)
                itr->label = inst.items[itr->label].id;    
        return AFGraph(NS.size()+1, At);
    }

    void solve(char vtype = 'I'){            
        GRBEnv* env = new GRBEnv();
        GRBModel model = GRBModel(*env);        
        model.set(GRB_StringAttr_ModelName, "flow");
        
        model.getEnv().set(GRB_IntParam_OutputFlag, 1);
        model.getEnv().set(GRB_IntParam_Threads, 1);
        model.getEnv().set(GRB_IntParam_Presolve, 1); 
        model.getEnv().set(GRB_IntParam_Method, 0); // primal simplex       
        model.getEnv().set(GRB_IntParam_MIPFocus, 1);
        model.getEnv().set(GRB_DoubleParam_Heuristics, 1);
        model.getEnv().set(GRB_DoubleParam_MIPGap, 0);
        model.getEnv().set(GRB_DoubleParam_MIPGapAbs, 1-1e-5);    
        
        
        int iS = 0;
        int iT = NS.size();
        
        sort(All(A));    
        map<Arc, GRBVar> va;    
        for(int i = 0; i < 3; i++){
            ForEach(a, A){
                if(i == 1 && a->u != iS) continue;
                if(i == 2 && a->v != iT) continue;
                if(i == 0 && (a->u == iS || a->v == iT)) continue;
                
                va[*a] = model.addVar(0.0, GRB_INFINITY, 0, vtype);
            }
        }
        GRBVar z = model.addVar(0.0, GRB_INFINITY, 1, vtype);        
        model.update();
        
        map<int, vector<Arc> > Al;
        map<int, vector<Arc> > in;
        map<int, vector<Arc> > out;
        ForEach(itr, A){
            if(itr->label != m)
                Al[itr->label].push_back(*itr);
            out[itr->u].push_back(*itr);
            in[itr->v].push_back(*itr);
        }
        
        ForEach(itr, Al){
            int dem = inst.items[itr->first].b;
            GRBLinExpr lin = 0;
            ForEach(a, itr->second) lin += va[*a];
            model.addConstr(lin >= dem);
        }
        
        for(int u = 0; u <= NS.size(); u++){
            GRBLinExpr lin = 0;
            ForEach(a, in[u]) lin += va[*a];
            ForEach(a, out[u]) lin -= va[*a];
            if(u == iS)
                model.addConstr(lin == -z);
            else if(u == iT)
                model.addConstr(lin == z);
            else
                model.addConstr(lin == 0);
        }
            
        model.optimize();                               

        printf("Total run time: %.3f seconds\n", TIMEDIF(tstart));

        if(vtype == 'I'){ 
            map<Arc, int> flow;              
            ForEach(a, va){
                double x = a->second.get(GRB_DoubleAttr_X);
                int rx = (int)round(x);
                assert(x - rx <= EPS);
                if(rx > 0){
                    int u = a->first.u;
                    int v = a->first.v;
                    int lbl = a->first.label;
                    Arc a(u, v, lbl);
                    flow[a] = rx;
                }
            }   
            
            ArcflowSol sol(flow, iS, iT);
            sol.print_solution(inst, false);           
        }
        
        free(env);
    }
};

#endif
