/**
This code is part of the Arc-flow 0-1CSP Solver.

Copyright (C) 2013, Filipe Brandao
Faculdade de Ciencias, Universidade do Porto
Porto, Portugal. All rights reserved. E-mail: <fdabrandao@dcc.fc.up.pt>.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#include "csp01flow.hpp"
#include <cstdio>
#include <cassert>
#include <cstring>
using namespace std;

#define EPSILON 1e-6

typedef vector<vector<int_pair> > adj_list;

vector<int> knapsack(const adj_list &adj, const vector<double> &values){
    const int nv = adj.size(), m = values.size();
    const int S = 0; 
    int best = S;
    vector<pair<double, int> > dp(nv, MP(-1, 0));    
    vector<int> pre_v(nv, -1);
    vector<int> pre_it(nv, -1);    
    dp[S] = MP(0, 0);
    for(int v = 0; v < nv; v++){
        ForEach(itr, adj[v]){
            const int &u = itr->first;
            const int &it = itr->second;
            pair<double, int> value = dp[u];
            if(it != m){
                value.first += values[it];
                value.second += 1;
            }
            if(value > dp[v]){
                dp[v] = value;
                pre_v[v] = u;
                pre_it[v] = it;
                if(value > dp[best])
                    best = v;
            }
        }
    }
    vector<int> result;
    if(dp[best].first > 1+EPSILON){
        int u = best;
        while(u != S){
            if(pre_it[u] != m)
                result.push_back(pre_it[u]);
            u = pre_v[u];    
        }
    }
    return result;
}

vector<int> knapsack0(const adj_list &adj, const vector<double> &values){
    const int nv = adj.size(), m = values.size();
    const int S = 0; 
    int best = S;
    vector<double> dp(nv, -1);    
    vector<int> pre_v(nv, -1);
    vector<int> pre_it(nv, -1);    
    dp[S] = 0;
    for(int v = 0; v < nv; v++){
        ForEach(itr, adj[v]){
            const int &u = itr->first;
            const int &it = itr->second;
            double value = dp[u];
            if(it != m)
                value += values[it];
            if(value > dp[v]){
                dp[v] = value;
                pre_v[v] = u;
                pre_it[v] = it;
                if(value > dp[best])
                    best = v;
            }
        }
    }
    vector<int> result;
    if(dp[best] > 1+EPSILON){
        int u = best;
        while(u != S){
            if(pre_it[u] != m)
                result.push_back(pre_it[u]);
            u = pre_v[u];    
        }
    }
    return result;
}

void solve(Instance inst){    
    clock_t t0 = CURTIME;
    
    const vector<Item> &items = inst.items;
    Arcflow graph(inst);
    AFGraph g = graph.get_graph();
    adj_list adj = g.get_adj(TRANSPOSE);        
    
    GRBEnv* env = new GRBEnv();
    GRBModel master = GRBModel(*env);    
    master.set(GRB_StringAttr_ModelName, "CSP");
    master.getEnv().set(GRB_IntParam_OutputFlag, 0);
    master.getEnv().set(GRB_IntParam_Threads, 1);          
    master.getEnv().set(GRB_IntParam_Method, 0); 

    GRBConstr rows[inst.m];
    for(int i = 0; i < inst.m; i++){
        GRBLinExpr lin = 0;
        rows[i] = master.addConstr(lin >= items[i].b);
    }
    master.update();
        
    for(int i = 0; i < inst.m; i++){        
        GRBColumn col = GRBColumn();        
        col.addTerm(1, rows[i]);
        master.addVar(0, GRB_INFINITY, 1, GRB_CONTINUOUS, col);                     
    }                    
    
    printf("m: %d\n", inst.m);
    vector<int> w(inst.m);
    for(int i = 0; i < inst.m; i++) w[i] = inst.items[i].w;
    vector<double> values(inst.m);
    for(int itr = inst.m; ; itr++){
        master.optimize();
        printf("%d: %.6f (%.3fs)\n", itr, master.get(GRB_DoubleAttr_ObjVal), TIMEDIF(t0));        
        for(int i = 0; i < inst.m; i++)
            values[i] = rows[i].get(GRB_DoubleAttr_Pi);            

        vector<int> sol = knapsack(adj, values);
        if(sol.empty()) break;
        
        GRBColumn col = GRBColumn();        
        ForEach(it, sol) col.addTerm(1, rows[*it]);
        master.addVar(0, GRB_INFINITY, 1, GRB_CONTINUOUS, col);
        master.update();
    }
        
    printf("zlp: %.6f\n", master.get(GRB_DoubleAttr_ObjVal));
    printf("nvars: %d\n", master.get(GRB_IntAttr_NumVars));
    printf("time: %.3f\n", TIMEDIF(t0));
    
    free(env);
}

int main(int argc, char *argv[]){     
    setvbuf(stdout, NULL, _IONBF, 0);    
    
    assert(argc >= 2 && argc <= 3);    
    
    FILE *fin = fopen(argv[1], "r");
    assert(fin != NULL);
         
    Instance inst;
    assert(fscanf(fin, "%d", &inst.W)==1);
    assert(fscanf(fin, "%d", &inst.m)==1);
    
    inst.items.resize(inst.m);
    for(int i = 0; i < inst.m; i++)
        assert(fscanf(fin, "%d%d", &inst.items[i].w, &inst.items[i].b)==2);    
    
    solve(inst);    

    fclose(fin);
    return 0;
}


