/**
This code is part of the Arc-flow 0-1CSP Solver.

Copyright (C) 2013, Filipe Brandao
Faculdade de Ciencias, Universidade do Porto
Porto, Portugal. All rights reserved. E-mail: <fdabrandao@dcc.fc.up.pt>.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#include <cstdio>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include "gurobi_c.h"
#include "gurobi_c++.h"
using namespace std;

#define EPSILON 1e-6
#define MAX_W 300
#define MAX_M 100
#define All(x) (x).begin(),(x).end()
#define ForEach(i,c) for(__typeof((c).begin()) i = (c).begin(); i != (c).end(); i++)
#define CURTIME clock()
#define TIMEDIF(t0) (double(clock()-t0)/CLOCKS_PER_SEC)

class Item{
public:
    int w;
    int b;
    int id;
    bool operator<(const Item &o) const{
        return w < o.w || (w == o.w && b < o.b);
    }
};

class Instance{
public:
    int W, m;
    vector<Item> items;
};

vector<int> knapsack(const vector<int> &w, const vector<double> &values, int W){
    static double DP[MAX_M+1][MAX_W+1];
    static int BEST[MAX_M+1][MAX_W+1];
    memset(DP, 0, sizeof(DP));
    memset(BEST, 0, sizeof(BEST));    
    int m = values.size();
    for(int i = 0; i < m; i++){
        for(int j = 0; j < w[i]; j++){
            DP[i+1][j] = DP[i][j];
            BEST[i+1][j] = j;
        }
        for(int j = w[i]; j <= W; j++){
            double a = DP[i][j-w[i]]+values[i];
            double b = DP[i][j];
            if(a >= b){
                DP[i+1][j] = a;
                BEST[i+1][j] = j-w[i];                                                
            }else{
                DP[i+1][j] = b;
                BEST[i+1][j] = j;
            }
        }
    }
    int u = W;
    vector<int> result;
    if(DP[m][u] > 1+EPSILON){
        for(int i = m; i > 0; i--){
            int v = BEST[i][u];
            if(v != u)
                result.push_back(i-1);
            u = v;
        }
    }
    return result;
}

void solve(Instance inst){    
    clock_t t0 = CURTIME;
    
    const vector<Item> &items = inst.items;   
    
    GRBEnv* env = new GRBEnv();
    GRBModel master = GRBModel(*env);    
    master.set(GRB_StringAttr_ModelName, "CSP");
    master.getEnv().set(GRB_IntParam_OutputFlag, 0);
    master.getEnv().set(GRB_IntParam_Threads, 1);          
    master.getEnv().set(GRB_IntParam_Method, 0); 

    GRBConstr rows[inst.m];
    for(int i = 0; i < inst.m; i++){
        GRBLinExpr lin = 0;
        rows[i] = master.addConstr(lin >= items[i].b);
    }
    master.update();
        
    for(int i = 0; i < inst.m; i++){        
        GRBColumn col = GRBColumn();        
        col.addTerm(1, rows[i]);
        master.addVar(0, GRB_INFINITY, 1, GRB_CONTINUOUS, col);                     
    }                    
    
    printf("m: %d\n", inst.m);
    vector<int> w(inst.m);
    for(int i = 0; i < inst.m; i++) w[i] = inst.items[i].w;
    vector<double> values(inst.m);
    for(int itr = inst.m; ; itr++){
        master.optimize();
        printf("%d: %.6f (%.3fs)\n", itr, master.get(GRB_DoubleAttr_ObjVal), TIMEDIF(t0));        
        for(int i = 0; i < inst.m; i++)
            values[i] = rows[i].get(GRB_DoubleAttr_Pi);            
            
        vector<int> sol = knapsack(w, values, inst.W);
        if(sol.empty()) break;
        
        GRBColumn col = GRBColumn();        
        ForEach(it, sol) col.addTerm(1, rows[*it]);
        master.addVar(0, GRB_INFINITY, 1, GRB_CONTINUOUS, col);
        master.update();
    }
        
    printf("zlp: %.6f\n", master.get(GRB_DoubleAttr_ObjVal));
    printf("nvars: %d\n", master.get(GRB_IntAttr_NumVars));
    printf("time: %.3f\n", TIMEDIF(t0));
    
    free(env);
}

int main(int argc, char *argv[]){     
    setvbuf(stdout, NULL, _IONBF, 0);    
    
    assert(argc >= 2 && argc <= 3);    
    
    FILE *fin = fopen(argv[1], "r");
    assert(fin != NULL);
         
    Instance inst;
    assert(fscanf(fin, "%d", &inst.W)==1);
    assert(fscanf(fin, "%d", &inst.m)==1);
    
    inst.items.resize(inst.m);
    for(int i = 0; i < inst.m; i++)
        assert(fscanf(fin, "%d%d", &inst.items[i].w, &inst.items[i].b)==2);    
    
    solve(inst);    

    fclose(fin);
    return 0;
}


