/**
This code is part of the Arc-flow 0-1CSP Solver.

Copyright (C) 2013, Filipe Brandao
Faculdade de Ciencias, Universidade do Porto
Porto, Portugal. All rights reserved. E-mail: <fdabrandao@dcc.fc.up.pt>.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#include <cstdio>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include "gurobi_c.h"
#include "gurobi_c++.h"
using namespace std;

#define EPSILON 1e-6
#define All(x) (x).begin(),(x).end()
#define ForEach(i,c) for(__typeof((c).begin()) i = (c).begin(); i != (c).end(); i++)
#define CURTIME clock()
#define TIMEDIF(t0) (double(clock()-t0)/CLOCKS_PER_SEC)

class Item{
public:
    int w;
    int b;
    int id;
    bool operator<(const Item &o) const{
        return w < o.w || (w == o.w && b < o.b);
    }
};

class Instance{
public:
    int W, m;
    vector<Item> items;
};

void solve(Instance inst){    
    clock_t t0 = CURTIME;
    
    const vector<Item> &items = inst.items;   
    
    GRBEnv* env1 = new GRBEnv();
    GRBModel master = GRBModel(*env1);    
    master.set(GRB_StringAttr_ModelName, "01CSP");
    master.getEnv().set(GRB_IntParam_OutputFlag, 0);
    master.getEnv().set(GRB_IntParam_Threads, 1);          
    master.getEnv().set(GRB_IntParam_Method, 0); 

    GRBEnv* env2 = new GRBEnv();
    GRBModel knapsack = GRBModel(*env2);  
    master.set(GRB_StringAttr_ModelName, "knapsack");    
    knapsack.getEnv().set(GRB_IntParam_OutputFlag, 0);
    knapsack.getEnv().set(GRB_IntParam_Threads, 1);   
    //knapsack.getEnv().set(GRB_IntParam_MIPFocus, 1);
    //knapsack.getEnv().set(GRB_DoubleParam_Heuristics, 0);    
    knapsack.set(GRB_IntAttr_ModelSense, GRB_MAXIMIZE);

    GRBConstr rows[inst.m];
    for(int i = 0; i < inst.m; i++){
        GRBLinExpr lin = 0;
        rows[i] = master.addConstr(lin >= items[i].b);
    }
    master.update();
        
    for(int i = 0; i < inst.m; i++){        
        GRBColumn col = GRBColumn();        
        col.addTerm(1, rows[i]);
        master.addVar(0, GRB_INFINITY, 1, GRB_CONTINUOUS, col);                     
    }                    
    
    GRBVar x[inst.m];
    for(int i = 0; i < inst.m; i++)
        x[i] = knapsack.addVar(0, 1, 0, GRB_BINARY);    
    knapsack.update();                 
    
    GRBLinExpr lin = 0;
    for(int i = 0; i < inst.m; i++)
        lin += items[i].w * x[i];
    knapsack.addConstr(lin <= inst.W);
    
    printf("m: %d\n", inst.m);
    for(int itr = inst.m; ; itr++){
        master.optimize();
        printf("%d: %.6f (%.3fs)\n", itr, master.get(GRB_DoubleAttr_ObjVal), TIMEDIF(t0));        
        for(int i = 0; i < inst.m; i++)
            x[i].set(GRB_DoubleAttr_Obj, rows[i].get(GRB_DoubleAttr_Pi));

        knapsack.optimize();
        if(knapsack.get(GRB_DoubleAttr_ObjVal) < 1+EPSILON) break;
                
        GRBColumn col = GRBColumn();        
        for(int i = 0; i < inst.m; i++){
            double val = x[i].get(GRB_DoubleAttr_X);
            if(val > EPSILON)
                col.addTerm(val, rows[i]);
        }
        master.addVar(0, GRB_INFINITY, 1, GRB_CONTINUOUS, col);
        master.update();
    }
        
    printf("zlp: %.6f\n", master.get(GRB_DoubleAttr_ObjVal));
    printf("nvars: %d\n", master.get(GRB_IntAttr_NumVars));
    printf("time: %.3f\n", TIMEDIF(t0));
    
    free(env1);
    free(env2);    
}

int main(int argc, char *argv[]){     
    setvbuf(stdout, NULL, _IONBF, 0);    
    
    assert(argc >= 2 && argc <= 3);    
    
    FILE *fin = fopen(argv[1], "r");
    assert(fin != NULL);
         
    Instance inst;
    assert(fscanf(fin, "%d", &inst.W)==1);
    assert(fscanf(fin, "%d", &inst.m)==1);
    
    inst.items.resize(inst.m);
    for(int i = 0; i < inst.m; i++)
        assert(fscanf(fin, "%d%d", &inst.items[i].w, &inst.items[i].b)==2);    
    
    solve(inst);    

    fclose(fin);
    return 0;
}


